
import torch
from torch.autograd     import grad
from shapes_toolbox.fidelities import kernel_distance, sinkhorn_distance, wasserstein_distance
from hausdorff    import hausdorff

def extract_point_cloud(I, affine) :
    """Bitmap to point cloud."""

    # Threshold, to extract the relevant indices ---------------------------------------
    ind = (I > .001).nonzero()

    # Extract the weights --------------------------------------------------------------
    D = len(I.shape)
    if   D == 2 : mu_i = I[ind[:,0], ind[:,1]]
    elif D == 3 : mu_i = I[ind[:,0], ind[:,1], ind[:,2]]
    else : raise NotImplementedError()

    mu_i = mu_i * affine[0,0] * affine[1,1] 

    # Don't forget the changes of coordinates! -----------------------------------------
    M   = affine[:D,:D] ; off = affine[:D,D]
    x_i = ind.float() @ M.t() + off

    #print("Warning !")
    #mu_i = mu_i/mu_i.sum()

    return ind, mu_i, x_i

def sparse_distance_bmp(params, A, B, affine_A, affine_B, normalize=False, info=False, action="measure") :
    """
    Takes as input two torch bitmaps (Tensors). 
    Returns a cost and a gradient, encoded as a vector bitmap.

    Args :
        - params, a dict which encodes.
        - A and B : two torch bitmaps (Tensors) of dimension D.
        - affine_A and affine_B : two matrices of size (D+1,D+1) (Tensors).
        - info : [True] or False; shall we sample the distance fields on the ambient space?
    """
    D = len(A.shape) # dimension of the ambient space, =2 for slices or =3 for volumes

    ind_A, alpha_i, x_i = extract_point_cloud(A, affine_A)
    ind_B, beta_j,  y_j = extract_point_cloud(B, affine_B)

    if normalize :
        alpha_i = alpha_i / alpha_i.sum()
        beta_j  = beta_j  / beta_j.sum()

    x_i.requires_grad = True
    if action == "image" :
        alpha_i.requires_grad = True

    # Compute the distance between the *measures* A and B ------------------------------
    print("{:,}-by-{:,} KP: ".format(len(x_i), len(y_j)), end='')

    routines = { "hausdorff" :    hausdorff ,
                 "sinkhorn"     : sinkhorn_distance,
                 "wasserstein"  : wasserstein_distance,
                 "kernel" :       kernel_distance    }
    routine = routines[ params.get("formula", "hausdorff") ]

    cost, heatmaps = routine( params, (alpha_i,x_i), (beta_j,y_j), info = info )
    if action == "image" :
        grad_a, grad_x = grad( cost, [alpha_i, x_i] ) # gradient wrt the voxels' positions and weights
    elif action == "measure" :
        grad_x = grad( cost, [x_i] )[0] # gradient wrt the voxels' positions

    # Point cloud to bitmap (grad_x) ---------------------------------------------------
    tensor   = torch.cuda.FloatTensor if A.is_cuda else torch.FloatTensor 
    # Using torch.zero(...).dtype(cuda.FloatTensor) would be inefficient...
    # Let's directly make a "malloc", before zero-ing in place
    grad_A = tensor( *(tuple(A.shape) + (D,))  )
    grad_A.zero_()

    if action == "image" :
        if   D == 2 :
            if True :
                dim_0 = affine_A[0,0] ; print(dim_0)
                grad_A[ind_A[:,0]  ,ind_A[:,1]  , :] += .25 * dim_0 * grad_x[:,:]
                grad_A[ind_A[:,0]+1,ind_A[:,1]  , :] += .25 * dim_0 * grad_x[:,:]
                grad_A[ind_A[:,0]  ,ind_A[:,1]+1, :] += .25 * dim_0 * grad_x[:,:]
                grad_A[ind_A[:,0]+1,ind_A[:,1]+1, :] += .25 * dim_0 * grad_x[:,:]

            grad_a = grad_a[:] * alpha_i[:]
            grad_A[ind_A[:,0]  ,ind_A[:,1]  , 0] -= .5*grad_a[:]
            grad_A[ind_A[:,0]+1,ind_A[:,1]  , 0] += .5*grad_a[:]
            grad_A[ind_A[:,0]  ,ind_A[:,1]+1, 0] -= .5*grad_a[:]
            grad_A[ind_A[:,0]+1,ind_A[:,1]+1, 0] += .5*grad_a[:]

            grad_A[ind_A[:,0]  ,ind_A[:,1]  , 1] -= .5*grad_a[:]
            grad_A[ind_A[:,0]  ,ind_A[:,1]+1, 1] += .5*grad_a[:]
            grad_A[ind_A[:,0]+1,ind_A[:,1]  , 1] -= .5*grad_a[:]
            grad_A[ind_A[:,0]+1,ind_A[:,1]+1, 1] += .5*grad_a[:]
 
            if False :
                grad_A[ind_A[:,0]  ,ind_A[:,1]  , 0] = grad_a[:]
                grad_A[ind_A[:,0]  ,ind_A[:,1]  , 1] = grad_a[:]
            
    elif action == "measure" :
        if   D == 2 : grad_A[ind_A[:,0],ind_A[:,1],            :] = grad_x[:,:] 
        elif D == 3 : grad_A[ind_A[:,0],ind_A[:,1],ind_A[:,2], :] = grad_x[:,:]
        else :        raise NotImplementedError()
    # N.B.: we return "PLUS gradient", i.e. "MINUS a descent direction".
    return cost, grad_A.detach(), heatmaps

